@@ -1,5 +1,4 @@
 from .codeg import (  # noqa: F401; Functions,
-    Attribute,
     BaseIndentPiece,
     BasePiece,
     ClassBlock,
@@ -8,6 +7,7 @@ from .codeg import (  # noqa: F401; Functions,
     FunctionBlock,
     GenericBlock,
     If,
+    Parameter,
     block,
     cls,
     comment,
@@ -19,6 +19,9 @@ from .codeg import (  # noqa: F401; Functions,
     import_,
     line,
     method,
+    p,
+    param,
+    parameter,
     script,
     try_,
     while_,
